@@ -1,5 +1,5 @@
 '''
-Copyright (c) 2012-2013, Agora Games, LLC All rights reserved.
+Copyright (c) 2012-2014, Agora Games, LLC All rights reserved.
 
 https://github.com/agoragames/kairos/blob/master/LICENSE.txt
 '''
@@ -87,32 +87,52 @@ class RedisBackend(Timeseries):
         rval.pop(interval)
 
     return rval
-  
-  def _insert(self, name, value, timestamp, intervals):
+
+  def _batch_insert(self, inserts, intervals, **kwargs):
+    '''
+    Specialized batch insert
+    '''
+    if 'pipeline' in kwargs:
+      pipe = kwargs.get('pipeline')
+      own_pipe = False
+    else:
+      pipe = self._client.pipeline(transaction=False)
+      kwargs['pipeline'] = pipe
+      own_pipe = True
+
+    for timestamp,names in inserts.iteritems():
+      for name,values in names.iteritems():
+        for value in values:
+          # TODO: support config param to flush the pipe every X inserts
+          self._insert( name, value, timestamp, intervals, **kwargs )
+
+    if own_pipe:
+      kwargs['pipeline'].execute()
+
+  def _insert(self, name, value, timestamp, intervals, **kwargs):
     '''
     Insert the value.
     '''
-    pipe = self._client.pipeline(transaction=False)
-    # TODO: apply the prefix if we're using one.
+    if 'pipeline' in kwargs:
+      pipe = kwargs.get('pipeline')
+    else:
+      pipe = self._client.pipeline(transaction=False)
 
     for interval,config in self._intervals.iteritems():
-      self._insert_data(name, value, timestamp, interval, config, pipe)
-      steps = intervals
-      if steps<0:
-        while steps<0:
-          i_timestamp = config['i_calc'].normalize(timestamp, steps)
-          self._insert_data(name, value, i_timestamp, interval, config, pipe)
-          steps += 1
-      elif steps>0:
-        while steps>0:
-          i_timestamp = config['i_calc'].normalize(timestamp, steps)
-          self._insert_data(name, value, i_timestamp, interval, config, pipe)
-          steps -= 1
+      timestamps = self._normalize_timestamps(timestamp, intervals, config)
+      for tstamp in timestamps:
+        self._insert_data(name, value, tstamp, interval, config, pipe)
 
-    pipe.execute()
+    if 'pipeline' not in kwargs:
+      pipe.execute()
 
   def _insert_data(self, name, value, timestamp, interval, config, pipe):
     '''Helper to insert data into redis'''
+    # Calculate the TTL and abort if inserting into the past
+    expire, ttl = config['expire'], config['ttl'](timestamp)
+    if expire and not ttl:
+      return
+
     i_bucket, r_bucket, i_key, r_key = self._calc_keys(config, name, timestamp)
 
     if config['coarse']:
@@ -125,11 +145,10 @@ class RedisBackend(Timeseries):
       pipe.sadd(i_key, r_bucket)
       self._type_insert(pipe, r_key, value)
 
-    expire = config['expire']
     if expire:
-      pipe.expire(i_key, expire)
+      pipe.expire(i_key, ttl)
       if not config['coarse']:
-        pipe.expire(r_key, expire)
+        pipe.expire(r_key, ttl)
 
   def delete(self, name):
     '''
@@ -153,8 +172,8 @@ class RedisBackend(Timeseries):
     i_bucket, r_bucket, i_key, r_key = self._calc_keys(config, name, timestamp)
     fetch = kws.get('fetch') or self._type_get
     process_row = kws.get('process_row') or self._process_row
-    
-    rval = OrderedDict()    
+
+    rval = OrderedDict()
     if config['coarse']:
       data = process_row( fetch(self._client, i_key) )
       rval[ config['i_calc'].from_bucket(i_bucket) ] = data
@@ -213,7 +232,7 @@ class RedisBackend(Timeseries):
           # TODO: use closures on the config for generating this resolution key
           resolution_key = '%s:%s'%(interval_key, bucket)
           fetch(pipe, resolution_key)
-        
+
         resolution_res = pipe.execute()
         for x,data in enumerate(resolution_res):
           i_t = config['i_calc'].from_bucket(interval_bucket)
@@ -258,18 +277,18 @@ class RedisCount(RedisBackend, Count):
         handle.incrbyfloat(key, value)
       else:
         handle.incr(key,value)
-  
+
   def _type_get(self, handle, key):
     return handle.get(key)
 
 class RedisGauge(RedisBackend, Gauge):
-  
+
   def _type_insert(self, handle, key, value):
     '''
     Insert the value into the series.
     '''
     handle.set(key, value)
-  
+
   def _type_get(self, handle, key):
     return handle.get(key)
 
